"""
Test that /user endpoints use the correct format.
"""
import unittest
from unittest.mock import patch
from storedsafe import StoredSafe
# pylint: disable=unused-wildcard-import,wildcard-import
from mocks import *


class User(unittest.TestCase):
    """Test /user endpoint"""

    def setUp(self):
        self.api = StoredSafe(
            host=MOCK_HOST, token=MOCK_TOKEN, **MOCK_DEFAULT_OPTIONS)

    @patch('requests.get', list_users)
    def test_list_users(self):
        """Successfully list users"""
        res = self.api.list_users(requests_options=MOCK_OVERRIDE_OPTIONS)
        self.assertEqual(res.status_code, 200)
        self.assertTrue(has_merged_options(res.options), True)

    @patch('requests.get', list_users_search_string)
    def test_list_users_search_string(self):
        """Successfully list users"""
        res = self.api.list_users(
            search_string=MOCK_SEARCH_STRING, requests_options=MOCK_OVERRIDE_OPTIONS)
        self.assertEqual(res.status_code, 200)
        self.assertTrue(has_merged_options(res.options), True)

    @patch('requests.get', get_user)
    def test_get_user(self):
        """Successfully get user"""
        res = self.api.get_user(
            MOCK_USER_ID, requests_options=MOCK_OVERRIDE_OPTIONS)
        self.assertEqual(res.status_code, 200)
        self.assertTrue(has_merged_options(res.options), True)

    @patch('requests.post', create_user)
    def test_create_user(self):
        """Successfully create user"""
        res = self.api.create_user(
            **MOCK_PARAMS, requests_options=MOCK_OVERRIDE_OPTIONS)
        self.assertEqual(res.status_code, 200)
        self.assertTrue(has_merged_options(res.options), True)

    @patch('requests.put', edit_user)
    def test_edit_user(self):
        """Successfully edit user"""
        res = self.api.edit_user(
            MOCK_USER_ID, **MOCK_PARAMS, requests_options=MOCK_OVERRIDE_OPTIONS)
        self.assertEqual(res.status_code, 200)
        self.assertTrue(has_merged_options(res.options), True)

    @patch('requests.delete', delete_user)
    def test_delete_user(self):
        """Successfully delete user"""
        res = self.api.delete_user(
            MOCK_USER_ID, requests_options=MOCK_OVERRIDE_OPTIONS)
        self.assertEqual(res.status_code, 200)
        self.assertTrue(has_merged_options(res.options), True)



"""
/vault endpoint mocks
"""
# pylint: disable=unused-wildcard-import,wildcard-import
from .mock_response import *


def list_vaults(*args, **kwargs):
    """Send mocked response for listing vaults."""
    if is_endpoint('/api/1.0/vault', args[0]):
        if has_valid_token(**kwargs):
            return MockSuccess(**kwargs)
        return MockError(**kwargs)
    return MockNotFound(**kwargs)


def vault_objects(*args, **kwargs):
    """Send mocked response for listing vault objects."""
    if is_endpoint(f'/api/1.0/vault/{MOCK_VAULT_ID}', args[0]):
        if has_valid_token(**kwargs):
            return MockSuccess(**kwargs)
        return MockError(**kwargs)
    return MockNotFound(**kwargs)


def vault_members(*args, **kwargs):
    """Send mocked response for listing vault objects."""
    if is_endpoint(f'/api/1.0/vault/{MOCK_VAULT_ID}/members', args[0]):
        if has_valid_token(**kwargs):
            return MockSuccess(**kwargs)
        return MockError(**kwargs)
    return MockNotFound(**kwargs)


def add_vault_member(*args, **kwargs):
    """Send mocked response for adding a member to a vault."""
    data = kwargs.get('json', {})
    if is_endpoint(f'/api/1.0/vault/{MOCK_VAULT_ID}/member/{MOCK_USER_ID}', args[0]):
        if has_valid_token(**kwargs) and data.get('status') == MOCK_STATUS:
            return MockSuccess(**kwargs)
        return MockError(**kwargs)
    return MockNotFound(**kwargs)


def edit_vault_member(*args, **kwargs):
    """Send mocked response for editing a member in a vault."""
    data = kwargs.get('json', {})
    if is_endpoint(f'/api/1.0/vault/{MOCK_VAULT_ID}/member/{MOCK_USER_ID}', args[0]):
        if has_valid_token(**kwargs) and data.get('status') == MOCK_STATUS:
            return MockSuccess(**kwargs)
        return MockError(**kwargs)
    return MockNotFound(**kwargs)


def remove_vault_member(*args, **kwargs):
    """Send mocked response for removing a member from a vault."""
    if is_endpoint(f'/api/1.0/vault/{MOCK_VAULT_ID}/member/{MOCK_USER_ID}', args[0]):
        if has_valid_token(**kwargs):
            return MockSuccess(**kwargs)
        return MockError(**kwargs)
    return MockNotFound(**kwargs)


def create_vault(*args, **kwargs):
    """Send mocked response for creating a vault."""
    data = kwargs.get('json', {})
    if is_endpoint('/api/1.0/vault', args[0]):
        if has_valid_token(**kwargs) and data == MOCK_PARAMS:
            return MockSuccess(**kwargs)
        return MockError(**kwargs)
    return MockNotFound(**kwargs)


def edit_vault(*args, **kwargs):
    """Send mocked response for editing a vault."""
    data = kwargs.get('json', {})
    if is_endpoint(f'/api/1.0/vault/{MOCK_VAULT_ID}', args[0]):
        if has_valid_token(**kwargs) and data == MOCK_PARAMS:
            return MockSuccess(**kwargs)
        return MockError(**kwargs)
    return MockNotFound(**kwargs)


def delete_vault(*args, **kwargs):
    """Send mocked response for deleting a vault."""
    if is_endpoint(f'/api/1.0/vault/{MOCK_VAULT_ID}', args[0]):
        if has_valid_token(**kwargs):
            return MockSuccess(**kwargs)
        return MockError(**kwargs)
    return MockNotFound(**kwargs)



"""
Test that /vault endpoints use the correct format.
"""
import unittest
from unittest.mock import patch
from storedsafe import StoredSafe
# pylint: disable=unused-wildcard-import,wildcard-import
from mocks import *


class Vault(unittest.TestCase):
    """Test /vault endpoint"""

    def setUp(self):
        self.api = StoredSafe(
            host=MOCK_HOST, token=MOCK_TOKEN, **MOCK_DEFAULT_OPTIONS)

    @patch('requests.get', list_vaults)
    def test_list_vaults(self):
        """Successfully list vaults"""
        res = self.api.list_vaults(requests_options=MOCK_OVERRIDE_OPTIONS)
        self.assertEqual(res.status_code, 200)
        self.assertTrue(has_merged_options(res.options), True)

    @patch('requests.get', vault_objects)
    def test_vault_objects(self):
        """Successfully list objects in vault"""
        res = self.api.vault_objects(
            MOCK_VAULT_ID, requests_options=MOCK_OVERRIDE_OPTIONS)
        self.assertEqual(res.status_code, 200)
        self.assertTrue(has_merged_options(res.options), True)

    @patch('requests.get', vault_members)
    def test_vault_members(self):
        """Successfully list members in vault"""
        res = self.api.vault_members(
            MOCK_VAULT_ID, requests_options=MOCK_OVERRIDE_OPTIONS)
        self.assertEqual(res.status_code, 200)
        self.assertTrue(has_merged_options(res.options), True)

    @patch('requests.post', add_vault_member)
    def test_add_vault_member(self):
        """Successful add member to vault"""
        res = self.api.add_vault_member(
            MOCK_VAULT_ID, MOCK_USER_ID, MOCK_STATUS, requests_options=MOCK_OVERRIDE_OPTIONS)
        self.assertEqual(res.status_code, 200)
        self.assertTrue(has_merged_options(res.options), True)

    @patch('requests.put', edit_vault_member)
    def test_edit_vault_member(self):
        """Successful edit member in vault"""
        res = self.api.edit_vault_member(
            MOCK_VAULT_ID, MOCK_USER_ID, MOCK_STATUS, requests_options=MOCK_OVERRIDE_OPTIONS)
        self.assertEqual(res.status_code, 200)
        self.assertTrue(has_merged_options(res.options), True)

    @patch('requests.delete', remove_vault_member)
    def test_remove_vault_member(self):
        """Successful remove member in vault"""
        res = self.api.remove_vault_member(
            MOCK_VAULT_ID, MOCK_USER_ID, requests_options=MOCK_OVERRIDE_OPTIONS)
        self.assertEqual(res.status_code, 200)
        self.assertTrue(has_merged_options(res.options), True)

    @patch('requests.post', create_vault)
    def test_create_vault(self):
        """Successful create new vault"""
        res = self.api.create_vault(
            **MOCK_PARAMS, requests_options=MOCK_OVERRIDE_OPTIONS)
        self.assertEqual(res.status_code, 200)
        self.assertTrue(has_merged_options(res.options), True)

    @patch('requests.put', edit_vault)
    def test_edit_vault(self):
        """Successful edit vault"""
        res = self.api.edit_vault(
            MOCK_VAULT_ID, **MOCK_PARAMS, requests_options=MOCK_OVERRIDE_OPTIONS)
        self.assertEqual(res.status_code, 200)
        self.assertTrue(has_merged_options(res.options), True)

    @patch('requests.delete', delete_vault)
    def test_delete_vault(self):
        """Successful delete vault"""
        res = self.api.delete_vault(
            MOCK_VAULT_ID, requests_options=MOCK_OVERRIDE_OPTIONS)
        self.assertEqual(res.status_code, 200)
        self.assertTrue(has_merged_options(res.options), True)



"""
Test that /object endpoints use the correct format.
"""

import unittest
import tempfile
from unittest.mock import patch
from storedsafe import StoredSafe

# pylint: disable=unused-wildcard-import,wildcard-import
import mocks


class Object(unittest.TestCase):
    """Test /object endpoint"""

    def setUp(self):
        self.api = StoredSafe(
            host=mocks.MOCK_HOST, token=mocks.MOCK_TOKEN, **mocks.MOCK_DEFAULT_OPTIONS
        )

    @patch("requests.get", mocks.get_object)
    def test_get_object(self):
        """Successfully get object"""
        res = self.api.get_object(
            mocks.MOCK_OBJECT_ID, requests_options=mocks.MOCK_OVERRIDE_OPTIONS
        )
        self.assertEqual(res.status_code, 200)
        self.assertTrue(mocks.has_merged_options(res.options), True)

    @patch("requests.get", mocks.get_object_children)
    def test_get_object_children(self):
        """Successfully get object with children"""
        res = self.api.get_object(
            mocks.MOCK_OBJECT_ID, True, requests_options=mocks.MOCK_OVERRIDE_OPTIONS
        )
        self.assertEqual(res.status_code, 200)
        self.assertTrue(mocks.has_merged_options(res.options), True)

    @patch("requests.get", mocks.decrypt_object)
    def test_decrypt_object(self):
        """Successfully decrypt object"""
        res = self.api.decrypt_object(
            mocks.MOCK_OBJECT_ID, requests_options=mocks.MOCK_OVERRIDE_OPTIONS
        )
        self.assertEqual(res.status_code, 200)
        self.assertTrue(mocks.has_merged_options(res.options), True)

    @patch("requests.post", mocks.create_object)
    def test_create_object(self):
        """Successful create object"""
        res = self.api.create_object(
            **mocks.MOCK_PARAMS, requests_options=mocks.MOCK_OVERRIDE_OPTIONS
        )
        self.assertEqual(res.status_code, 200)
        self.assertTrue(mocks.has_merged_options(res.options), True)

    @patch("requests.put", mocks.edit_object)
    def test_edit_object(self):
        """Successful edit object"""
        res = self.api.edit_object(
            mocks.MOCK_OBJECT_ID,
            **mocks.MOCK_PARAMS,
            requests_options=mocks.MOCK_OVERRIDE_OPTIONS,
        )
        self.assertEqual(res.status_code, 200)
        self.assertTrue(mocks.has_merged_options(res.options), True)

    @patch("requests.delete", mocks.delete_object)
    def test_delete_object(self):
        """Successful delete object"""
        res = self.api.delete_object(
            mocks.MOCK_OBJECT_ID, requests_options=mocks.MOCK_OVERRIDE_OPTIONS
        )
        self.assertEqual(res.status_code, 200)
        self.assertTrue(mocks.has_merged_options(res.options), True)

    @patch("requests.get", mocks.find)
    def test_find(self):
        """Successfully request find"""
        res = self.api.find(
            mocks.MOCK_NEEDLE, requests_options=mocks.MOCK_OVERRIDE_OPTIONS
        )
        self.assertEqual(res.status_code, 200)
        self.assertTrue(mocks.has_merged_options(res.options), True)

    @patch("requests.get", mocks.get_file)
    def test_get_file(self):
        """Successfully request file"""
        res = self.api.get_file(
            mocks.MOCK_OBJECT_ID, requests_options=mocks.MOCK_OVERRIDE_OPTIONS
        )
        self.assertEqual(res.status_code, 200)
        self.assertTrue(mocks.has_merged_options(res.options), True)

    @patch("requests.post", mocks.get_mime_type)
    def test_get_mime_type(self):
        """Successfully request mime type"""
        with tempfile.NamedTemporaryFile(suffix=".txt") as fp:
            fp.write(mocks.MOCK_FILE_CONTENT)
            fp.flush()
            res = self.api.get_mime_type(fp.name, requests_options=mocks.MOCK_OVERRIDE_OPTIONS)
            self.assertEqual(res.status_code, 200)
            self.assertTrue(mocks.has_merged_options(res.options), True)

    @patch("requests.post", mocks.filecollect)
    def test_filecollect(self):
        """Successfully request filecollect"""
        with tempfile.NamedTemporaryFile(suffix=".txt") as fp:
            fp.write(mocks.MOCK_FILE_CONTENT)
            fp.flush()
            res = self.api.filecollect(fp.name, requests_options=mocks.MOCK_OVERRIDE_OPTIONS)
            self.assertEqual(res.status_code, 200)
            self.assertTrue(mocks.has_merged_options(res.options), True)

    @patch("requests.post", mocks.file_upload)
    def test_file_upload(self):
        """Successfully request file upload"""
        with tempfile.NamedTemporaryFile(suffix=".txt") as fp:
            fp.write(mocks.MOCK_FILE_CONTENT)
            fp.flush()
            res = self.api.upload_file(fp.name, **mocks.MOCK_PARAMS, requests_options=mocks.MOCK_OVERRIDE_OPTIONS)
            self.assertEqual(res.status_code, 200)
            self.assertTrue(mocks.has_merged_options(res.options), True)



"""
Test that /utils endpoints use the correct format.
"""
import unittest
from unittest.mock import patch
from storedsafe import StoredSafe
# pylint: disable=unused-wildcard-import,wildcard-import
from mocks import *


class Utils(unittest.TestCase):
    """Test /utils endpoint"""

    def setUp(self):
        self.api = StoredSafe(host=MOCK_HOST, token=MOCK_TOKEN, **MOCK_DEFAULT_OPTIONS)

    @patch('requests.get', status_values)
    def test_status_values(self):
        """Successfully list status values"""
        res = self.api.status_values(requests_options=MOCK_OVERRIDE_OPTIONS)
        self.assertEqual(res.status_code, 200)
        self.assertTrue(has_merged_options(res.options), True)

    @patch('requests.get', password_policies)
    def test_password_policies(self):
        """Successfully list password policies"""
        res = self.api.password_policies(requests_options=MOCK_OVERRIDE_OPTIONS)
        self.assertEqual(res.status_code, 200)
        self.assertTrue(has_merged_options(res.options), True)

    @patch('requests.get', version)
    def test_version(self):
        """Successfully get StoredSafe version"""
        res = self.api.version(requests_options=MOCK_OVERRIDE_OPTIONS)
        self.assertEqual(res.status_code, 200)
        self.assertTrue(has_merged_options(res.options), True)

    @patch('requests.get', generate_password)
    def test_generate_password(self):
        """Successfully generate password"""
        res = self.api.generate_password(requests_options=MOCK_OVERRIDE_OPTIONS)
        self.assertEqual(res.status_code, 200)
        self.assertTrue(has_merged_options(res.options), True)

    @patch('requests.get', generate_password_params)
    def test_generate_password_params(self):
        """Successfully generate password with parameters"""
        res = self.api.generate_password(**MOCK_PARAMS, requests_options=MOCK_OVERRIDE_OPTIONS)
        self.assertEqual(res.status_code, 200)
        self.assertTrue(has_merged_options(res.options), True)



"""
Test that /auth endpoints use the correct format.
"""
import unittest
from unittest.mock import patch
from storedsafe import StoredSafe
# pylint: disable=unused-wildcard-import,wildcard-import
from mocks import *


class Auth(unittest.TestCase):
    """Test /auth endpoint"""

    @patch('requests.post', login_totp)
    def test_login_totp(self):
        """Successful login using TOTP"""
        api = StoredSafe(host=MOCK_HOST, apikey=MOCK_APIKEY,
                         **MOCK_DEFAULT_OPTIONS)
        res = api.login_totp(MOCK_USERNAME, MOCK_PASSPHRASE,
                             MOCK_OTP, MOCK_OVERRIDE_OPTIONS)

        self.assertEqual(res.status_code, 200)
        self.assertEqual(api.token, MOCK_TOKEN)
        self.assertTrue(has_merged_options(res.options), True)

    @patch('requests.post', login_yubikey)
    def test_login_yubikey(self):
        """Successful login using YubiKey"""
        api = StoredSafe(host=MOCK_HOST, apikey=MOCK_APIKEY,
                         **MOCK_DEFAULT_OPTIONS)
        res = api.login_yubikey(
            MOCK_USERNAME, MOCK_PASSPHRASE, MOCK_OTP, MOCK_OVERRIDE_OPTIONS)

        self.assertEqual(res.status_code, 200)
        self.assertEqual(api.token, MOCK_TOKEN)
        self.assertTrue(has_merged_options(res.options), True)

    @patch('requests.post', login_smartcard)
    def test_login_smartcard(self):
        """Successful login using mTLS"""
        api = StoredSafe(host=MOCK_HOST, apikey=MOCK_APIKEY,
                         **MOCK_DEFAULT_OPTIONS)
        res = api.login_smartcard(
            MOCK_USERNAME, MOCK_PASSPHRASE, MOCK_PUBKEY, MOCK_PRIVKEY,
            MOCK_OVERRIDE_OPTIONS)

        self.assertEqual(res.status_code, 200)
        self.assertEqual(api.token, MOCK_TOKEN)
        self.assertTrue(has_merged_options(res.options), True)

    @patch('requests.get', logout)
    def test_logout(self):
        """Successful logout"""
        api = StoredSafe(host=MOCK_HOST, token=MOCK_TOKEN,
                         **MOCK_DEFAULT_OPTIONS)
        res = api.logout(MOCK_OVERRIDE_OPTIONS)
        self.assertEqual(res.status_code, 200)
        self.assertTrue(has_merged_options(res.options), True)

    @patch('requests.post', check)
    def test_check(self):
        """Successful check"""
        api = StoredSafe(host=MOCK_HOST, token=MOCK_TOKEN,
                         **MOCK_DEFAULT_OPTIONS)
        res = api.check(MOCK_OVERRIDE_OPTIONS)
        self.assertEqual(res.status_code, 200)
        self.assertTrue(has_merged_options(res.options), True)



"""
/auth endpoint mocks
"""
# pylint: disable=unused-wildcard-import,wildcard-import
from .mock_response import *


def login_totp(*args, **kwargs):
    """Send mocked response for totp login."""
    data = kwargs.get('json', {})
    if is_endpoint('/api/1.0/auth', args[0]):
        login_type = data.get('logintype', 'yubikey')
        if login_type == 'totp':
            if (
                    data.get('username') == MOCK_USERNAME and
                    data.get('passphrase') == MOCK_PASSPHRASE and
                    data.get('otp') == MOCK_OTP and
                    data.get('apikey') == MOCK_APIKEY
            ):
                return MockLoginSuccess(**kwargs)
        return MockError(**kwargs)
    return MockNotFound(**kwargs)


def login_yubikey(*args, **kwargs):
    """Send mocked response for yubikey login."""
    data = kwargs.get('json', {})
    if is_endpoint('/api/1.0/auth', args[0]):
        login_type = data.get('logintype', 'yubikey')
        if login_type == 'yubikey':
            if (
                    data.get('username') == MOCK_USERNAME and
                    data.get('keys') == MOCK_PASSPHRASE +
                MOCK_APIKEY + MOCK_OTP and
                    data.get('apikey') == MOCK_APIKEY
            ):
                return MockLoginSuccess(**kwargs)
        return MockError(**kwargs)
    return MockNotFound(**kwargs)


def login_smartcard(*args, **kwargs):
    """Send mocked response for smartcard login."""
    data = kwargs.get('json', {})
    if is_endpoint('/api/1.0/auth', args[0], mtls=True):
        login_type = data.get('logintype', 'yubikey')
        if login_type == 'smartcard':
            if (
                    data.get('username') == MOCK_USERNAME and
                    data.get('passphrase') == MOCK_PASSPHRASE and
                    kwargs.get('cert') == (MOCK_PUBKEY, MOCK_PRIVKEY) and
                    data.get('apikey') == MOCK_APIKEY
            ):
                return MockLoginSuccess(**kwargs)
        return MockError(**kwargs)
    return MockNotFound(**kwargs)


def logout(*args, **kwargs):
    """Send mocked response for logout."""
    if is_endpoint('/api/1.0/auth/logout', args[0]):
        if has_valid_token(**kwargs):
            return MockSuccess(**kwargs)
        return MockError(**kwargs)
    return MockNotFound(**kwargs)


def check(*args, **kwargs):
    """Send mocked response for check."""
    if is_endpoint('/api/1.0/auth/check', args[0]):
        if has_valid_token(**kwargs):
            return MockSuccess(**kwargs)
        return MockError(**kwargs)
    return MockNotFound(**kwargs)



"""
/object endpoint mocks
"""
# pylint: disable=unused-wildcard-import,wildcard-import
from .mock_response import *


def status_values(*args, **kwargs):
    """Send mocked response listing status values."""
    if is_endpoint('/api/1.0/utils/statusvalues', args[0]):
        if has_valid_token(**kwargs):
            return MockSuccess(**kwargs)
        return MockError(**kwargs)
    return MockNotFound(**kwargs)


def password_policies(*args, **kwargs):
    """Send mocked response listing password policies."""
    if is_endpoint('/api/1.0/utils/policies', args[0]):
        if has_valid_token(**kwargs):
            return MockSuccess(**kwargs)
        return MockError(**kwargs)
    return MockNotFound(**kwargs)


def version(*args, **kwargs):
    """Send mocked response getting StoredSafe version."""
    if is_endpoint('/api/1.0/utils/version', args[0]):
        if has_valid_token(**kwargs):
            return MockSuccess(**kwargs)
        return MockError(**kwargs)
    return MockNotFound(**kwargs)

def generate_password(*args, **kwargs):
    """Send mocked response generating password."""
    if is_endpoint('/api/1.0/utils/pwgen', args[0]):
        if has_valid_token(**kwargs):
            return MockSuccess(**kwargs)
        return MockError(**kwargs)
    return MockNotFound(**kwargs)

def generate_password_params(*args, **kwargs):
    """Send mocked response generating password with parameters."""
    data = kwargs.get('params', {})
    if is_endpoint('/api/1.0/utils/pwgen', args[0]):
        if has_valid_token(**kwargs) and data == MOCK_PARAMS:
            return MockSuccess(**kwargs)
        return MockError(**kwargs)
    return MockNotFound(**kwargs)



"""
/object endpoint mocks
"""
import mocks.mock_response as mockr
from base64 import b64encode


def get_object(*args, **kwargs):
    """Send mocked response getting object."""
    if mockr.is_endpoint(f"/api/1.0/object/{mockr.MOCK_OBJECT_ID}", args[0]):
        if mockr.has_valid_token(**kwargs):
            return mockr.MockSuccess(**kwargs)
        return mockr.MockError(**kwargs)
    return mockr.MockNotFound(**kwargs)


def get_object_children(*args, **kwargs):
    """Send mocked response getting object."""
    data = kwargs.get("params", {})
    if mockr.is_endpoint(f"/api/1.0/object/{mockr.MOCK_OBJECT_ID}", args[0]):
        if mockr.has_valid_token(**kwargs) and data.get("children") == "true":
            return mockr.MockSuccess(**kwargs)
        return mockr.MockError(**kwargs)
    return mockr.MockNotFound(**kwargs)


def decrypt_object(*args, **kwargs):
    """Send mocked response for decrypting object."""
    data = kwargs.get("params", {})
    if mockr.is_endpoint(f"/api/1.0/object/{mockr.MOCK_OBJECT_ID}", args[0]):
        if mockr.has_valid_token(**kwargs) and data.get("decrypt") == "true":
            return mockr.MockSuccess(**kwargs)
        return mockr.MockError(**kwargs)
    return mockr.MockNotFound(**kwargs)


def get_file(*args, **kwargs):
    """Send mocked response for getting file."""
    data = kwargs.get("params", {})
    if mockr.is_endpoint(f"/api/1.0/object/{mockr.MOCK_OBJECT_ID}", args[0]):
        if (
            mockr.has_valid_token(**kwargs)
            and data.get("decrypt") == "true"
            and data.get("filedata") == "true"
        ):
            return mockr.MockSuccess(**kwargs)
        return mockr.MockError(**kwargs)
    return mockr.MockNotFound(**kwargs)


def get_mime_type(*args, **kwargs):
    """Send mocked response for getting file mime type."""
    data = kwargs.get("json", {})
    if mockr.is_endpoint("/api/1.0/utils/get_mime_type", args[0]):
        if (
            mockr.has_valid_token(**kwargs)
            and data.get("size") == 128
            and data.get("extension") == mockr.MOCK_FILE_EXTENSION
            and data.get("data") == b64encode(mockr.MOCK_FILE_CONTENT[:64]).decode("utf-8")
        ):
            return mockr.MockSuccess(**kwargs)
        return mockr.MockError(**kwargs)
    return mockr.MockNotFound(**kwargs)

def filecollect(*args, **kwargs):
    """Send mocked response for filecollect."""
    files = kwargs.get("files", {})
    if mockr.is_endpoint("/api/1.0/filecollect", args[0]):
        if (
            mockr.has_valid_token(**kwargs)
            and files.get("upload")[0].endswith(mockr.MOCK_FILE_EXTENSION)
            and files.get("upload")[1] == mockr.MOCK_FILE_CONTENT[:64]
        ):
            return mockr.MockSuccess(**kwargs)
        return mockr.MockError(**kwargs)
    return mockr.MockNotFound(**kwargs)

def file_upload(*args, **kwargs):
    """Send mocked response for filecollect."""
    data = kwargs.get("data", {})
    files = kwargs.get("files", {})
    # Close file pointer since requests isn't doing it due to mock
    content = files.get('upload').read()
    files.get('upload').close()
    if mockr.is_endpoint("/api/1.0/object", args[0]):
        if (
            mockr.has_valid_token(**kwargs)
            and data == mockr.MOCK_FILE_PARAMS
            and files.get("upload").name.endswith(mockr.MOCK_FILE_EXTENSION)
            and files.get("upload").mode == "rb"
            and content == mockr.MOCK_FILE_CONTENT
        ):
            return mockr.MockSuccess(**kwargs)
        return mockr.MockError(**kwargs)
    return mockr.MockNotFound(**kwargs)


def create_object(*args, **kwargs):
    """Send mocked response for creating an object."""
    data = kwargs.get("json", {})
    if mockr.is_endpoint("/api/1.0/object", args[0]):
        if mockr.has_valid_token(**kwargs) and data == mockr.MOCK_PARAMS:
            return mockr.MockSuccess(**kwargs)
        return mockr.MockError(**kwargs)
    return mockr.MockNotFound(**kwargs)


def edit_object(*args, **kwargs):
    """Send mocked response for editing an object."""
    data = kwargs.get("json", {})
    if mockr.is_endpoint(f"/api/1.0/object/{mockr.MOCK_OBJECT_ID}", args[0]):
        if mockr.has_valid_token(**kwargs) and data == mockr.MOCK_PARAMS:
            return mockr.MockSuccess(**kwargs)
        return mockr.MockError(**kwargs)
    return mockr.MockNotFound(**kwargs)


def delete_object(*args, **kwargs):
    """Send mocked response for deleting an object."""
    if mockr.is_endpoint(f"/api/1.0/object/{mockr.MOCK_OBJECT_ID}", args[0]):
        if mockr.has_valid_token(**kwargs):
            return mockr.MockSuccess(**kwargs)
        return mockr.MockError(**kwargs)
    return mockr.MockNotFound(**kwargs)


def find(*args, **kwargs):
    """Send mocked response for finding objects."""
    data = kwargs.get("params", {})
    if mockr.is_endpoint("/api/1.0/find", args[0]):
        if mockr.has_valid_token(**kwargs) and data.get("needle") == mockr.MOCK_NEEDLE:
            return mockr.MockSuccess(**kwargs)
        return mockr.MockError(**kwargs)
    return mockr.MockNotFound(**kwargs)



"""
/object endpoint mocks
"""
# pylint: disable=unused-wildcard-import,wildcard-import
from .mock_response import *


def list_templates(*args, **kwargs):
    """Send mocked response listing all templates."""
    if is_endpoint(f'/api/1.0/template', args[0]):
        if has_valid_token(**kwargs):
            return MockSuccess(**kwargs)
        return MockError(**kwargs)
    return MockNotFound(**kwargs)


def get_template(*args, **kwargs):
    """Send mocked response getting template."""
    if is_endpoint(f'/api/1.0/template/{MOCK_TEMPLATE_ID}', args[0]):
        if has_valid_token(**kwargs):
            return MockSuccess(**kwargs)
        return MockError(**kwargs)
    return MockNotFound(**kwargs)



"""StoredSafe API wrapper module."""
from pathlib import Path
from base64 import b64encode
import requests


class RCException(Exception):
    """Failed to read rc file."""


class ApikeyUndefinedException(Exception):
    """Cannot authenticate with StoredSafe API calls because apikey is not defined."""


class TokenUndefinedException(Exception):
    """Cannot use privileged StoredSafe API calls because token is not defined."""


# pylint: disable=too-many-public-methods
class StoredSafe:
    """StoredSafe API wrapper class."""

    @staticmethod
    def from_rc(path=Path.home() / '.storedsafe-client.rc', **requests_options):
        """Create StoredSafe instance from rc-file"""
        config = {}
        try:
            with Path(path).open('r') as rc_file:
                for line in rc_file:
                    [key, value] = line.strip().split(':')
                    if key == 'mysite':
                        config['host'] = value
                    elif key == 'apikey':
                        config['apikey'] = value
                    elif key == 'token':
                        config['token'] = value
        except Exception:
            raise RCException()
        return StoredSafe(**config, **requests_options)

    def __init__(self, host, apikey=None, token=None, version='1.0', **requests_options):
        self.host = host
        self.apikey = apikey
        self.token = token
        self.api_version = version
        self.requests_options = requests_options

    ###
    # Helper methods.
    ##
    def __assert_apikey_exists(self):
        """Throws ApikeyUndefinedException if token is None."""
        if self.apikey is None:
            raise ApikeyUndefinedException()

    def __headers(self, requests_options, token=True):
        """Create the required headers for requests to the StoredSafe API and merge with options."""
        headers = {
            **self.requests_options.get('headers', {}),
            **requests_options.get('headers', {})}
        if token:
            self.__assert_token_exists()
            headers['X-Http-Token'] = self.token
        return headers

    def __auth(self, data, mtls=False, **requests_options):
        """Authenticate with StoredSafe and save token if the request was successful."""
        self.__assert_apikey_exists()
        res = requests.post(
            self.__get_url('/auth', mtls), **{
            **self.requests_options,
            **requests_options,
            'headers': self.__headers(requests_options, False)
            }, json=data)
        if res.status_code == 200:
            data = res.json()
            self.token = data['CALLINFO']['token']
        return res


    def __assert_token_exists(self):
        """Throws TokenUndefinedException if token is None."""
        if self.token is None:
            raise TokenUndefinedException()

    def __get_url(self, path, mtls=False):
        """Get the full url of the relative API path."""
        port = mtls and ':8443' or ''
        return f'https://{self.host}{port}/api/{self.api_version}/{path.strip("/")}'

    def __get(self, path, params=None, mtls=False, **requests_options):
        """Send a GET request to the provided relative API path."""
        self.__assert_token_exists()
        args = {
            **self.requests_options,
            **requests_options,
            'headers': self.__headers(requests_options)
        }
        if params is not None:
            args['params'] = params
        return requests.get(
            self.__get_url(path, mtls), **args)

    def __post(self, path, data={}, mtls=False, **requests_options):
        """Send a POST request to the provided relative API path."""
        self.__assert_token_exists()
        return requests.post(
            self.__get_url(path, mtls), **{
                **self.requests_options,
                **requests_options,
                'headers': self.__headers(requests_options)
            }, json=data)

    def __put(self, path, data={}, mtls=False, **requests_options):
        """Send a PUT request to the provided relative API path."""
        self.__assert_token_exists()
        return requests.put(
            self.__get_url(path, mtls), **{
                **self.requests_options,
                **requests_options,
                'headers': self.__headers(requests_options)
            }, json=data)

    def __delete(self, path, mtls=False, **requests_options):
        """Send a DELETE request to the provided relative API path."""
        self.__assert_token_exists()
        return requests.delete(
            self.__get_url(path, mtls), **{
                **self.requests_options,
                **requests_options,
                'headers': self.__headers(requests_options)
            })

    def __post_file(self, path, file, data={}, mtls=False, **requests_options):
        """Send a POST request to the provided relative API path."""
        self.__assert_token_exists()
        return requests.post(
            self.__get_url(path, mtls), **{
                **self.requests_options,
                **requests_options,
                'headers': self.__headers(requests_options)
            }, files={'upload': file}, data=data)

    ###
    # API Auth methods.
    ##
    def login_totp(self, username, passphrase, otp, requests_options={}):
        """Request login using TOTP."""
        data = {
            'username': username,
            'passphrase': passphrase,
            'otp': otp,
            'apikey': self.apikey,
            'logintype': 'totp',
        }
        return self.__auth(data, **requests_options)

    def login_yubikey(self, username, passphrase, otp, requests_options={}):
        """Request login using Yubico OTP."""
        data = {
            'username': username,
            'keys': f'{passphrase}{self.apikey}{otp}',
            'apikey': self.apikey,
        }
        return self.__auth(data, **requests_options)

    def login_smartcard(self, username, passphrase, cert, key, requests_options={}):
        """Request login using mTLS"""
        data = {
            'username': username,
            'passphrase': passphrase,
            'apikey': self.apikey,
            'logintype': 'smartcard',
        }
        return self.__auth(data, mtls=True, cert=(cert, key), **requests_options)

    def logout(self, requests_options={}):
        """Request logout."""
        return self.__get('/auth/logout', **requests_options)

    def check(self, requests_options={}):
        """Request check status of session."""
        return self.__post('/auth/check', **requests_options)

    ###
    # API Vault methods.
    ##
    def list_vaults(self, requests_options={}):
        """Request a list of all vaults."""
        return self.__get('/vault', **requests_options)

    def vault_objects(self, vault_id, requests_options={}):
        """Request a list of the objects in a vault."""
        return self.__get(f'/vault/{vault_id}', **requests_options)

    def vault_members(self, vault_id, requests_options={}):
        """Request a list of the members in a vault."""
        return self.__get(f'/vault/{vault_id}/members', **requests_options)

    def add_vault_member(self, vault_id, user_id, status, requests_options={}):
        """Request to add a member to a vault."""
        return self.__post(f'/vault/{vault_id}/member/{user_id}', {'status': status}, **requests_options)

    def edit_vault_member(self, vault_id, user_id, status, requests_options={}):
        """Request to edit a member in a vault."""
        return self.__put(f'/vault/{vault_id}/member/{user_id}', {'status': status}, **requests_options)

    def remove_vault_member(self, vault_id, user_id, requests_options={}):
        """Request to remove a member from a vault."""
        return self.__delete(f'/vault/{vault_id}/member/{user_id}', **requests_options)

    def create_vault(self, requests_options={}, **params):
        """Request the creation of a new vault."""
        return self.__post('/vault', params, **requests_options)

    def edit_vault(self, vault_id, requests_options={}, **params):
        """Request an edit of an existing vault."""
        return self.__put(f'/vault/{vault_id}', params, **requests_options)

    def delete_vault(self, vault_id, requests_options={}):
        """Request the deletion of a vault."""
        return self.__delete(f'/vault/{vault_id}', **requests_options)

    ###
    # API Object methods.
    ##
    def get_object(self, object_id, children=False, requests_options={}):
        """Request a StoredSafe object and optionally its children."""
        return self.__get(f'/object/{object_id}', {'children': 'true' if children else 'false'}, **requests_options)

    def decrypt_object(self, object_id, requests_options={}):
        """Request the decryption of a StoredSafe object."""
        return self.__get(f'/object/{object_id}', {'decrypt': 'true'}, **requests_options)

    def create_object(self, requests_options={}, **params):
        """Request the creation of an object."""
        return self.__post('/object', params, **requests_options)

    def edit_object(self, object_id, requests_options={}, **params):
        """Request the edit of an existing object."""
        return self.__put(f'/object/{object_id}', params, **requests_options)

    def delete_object(self, object_id, requests_options={}):
        """Request the deletion of an object."""
        return self.__delete(f'/object/{object_id}', **requests_options)

    def find(self, needle, requests_options={}):
        """Request all objects with searchable fields matching the needle."""
        return self.__get('/find', {'needle': needle}, **requests_options)

    ###
    # API File (Object) methods.
    ##
    def get_mime_type(self, file, requests_options={}, **params):
        """Check the mime type of the file and get the max upload size."""
        data = {}
        path = Path(file)
        data['extension'] = path.suffix
        data['size'] = path.stat().st_size
        with open(file, 'rb') as f:
            data['data'] = b64encode(f.read(64)).decode("utf-8")
        return self.__post('/utils/get_mime_type', { **data, **params }, **requests_options)

    def filecollect(self, file, requests_options={}, **params):
        """Request the appropriate template for the file."""
        path = Path(file)
        with path.open('rb') as f:
            data = f.read(64)
        return self.__post_file('/filecollect', (path.name, data), params, **requests_options)

    def upload_file(self, file, requests_options={}, **params):
        """Request the creation of a file object."""
        if not params.get('templateid'):
            params['templateid'] = 3
        return self.__post_file('/object', open(file, 'rb'), params, **requests_options)

    def get_file(self, object_id, requests_options={}):
        """Request a base64 string of a file object."""
        return self.__get(f'/object/{object_id}', {'decrypt': 'true', 'filedata': 'true'}, **requests_options)


    ###
    # API Templates methods.
    ##
    def list_templates(self, requests_options={}):
        """Request a list of all available templates."""
        return self.__get('/template', **requests_options)

    def get_template(self, template_id, requests_options={}):
        """Request a StoredSafe template."""
        return self.__get(f'/template/{template_id}', **requests_options)

    ###
    # API User methods
    ##
    def list_users(self, search_string=None, requests_options={}):
        """Request list of all users or any users matching search string."""
        if search_string is None:
            return self.__get('/user', **requests_options)
        return self.__get(f'/user/{search_string}', **requests_options)

    def get_user(self, user_id, requests_options={}):
        """Request user matching the user_id."""
        return self.__get(f'/user/{user_id}', **requests_options)

    def create_user(self, requests_options={}, **params):
        """Request the creation of a new user."""
        return self.__post('/user', params, **requests_options)

    def edit_user(self, user_id, requests_options={}, **params):
        """Request the creation of a new user."""
        return self.__put(f'/user/{user_id}', params, **requests_options)

    def delete_user(self, user_id, requests_options={}):
        """Request the creation of a new user."""
        return self.__delete(f'/user/{user_id}', **requests_options)

    def get_user_certificate(self, user_id, requests_options={}):
        """Request the user certificate used for mTLS"""
        return self.__get(f'/usercert/{user_id}', **requests_options)

    def set_user_certificate(self, user_id, cert_path, requests_options={}):
        """Request a new user certificate for mTLS to be added to the user"""
        files = {'user_cert': open(cert_path, 'rb')}
        return self.__post(f'/usercert/{user_id}', files=files, **requests_options)

    def remove_user_certificate(self, user_id, requests_options={}):
        """Request the user certificate for mTLS to be removed from user"""
        return self.__delete(f'/usercert/{user_id}', **requests_options)

    ###
    # API Utils methods.
    ##
    def status_values(self, requests_options={}):
        """Request a list of all available capabilities and permission bits."""
        return self.__get('/utils/statusvalues', **requests_options)

    def password_policies(self, requests_options={}):
        """Request a list of all available password policies."""
        return self.__get('/utils/policies', **requests_options)

    def version(self, requests_options={}):
        """Request the version of the StoredSafe server."""
        return self.__get('/utils/version', **requests_options)

    def generate_password(self, requests_options={}, **params):
        """Request a password generated with the passed settings (or vault policy)."""
        return self.__get('/utils/pwgen', params, **requests_options)



"""
Define mocked responses mimicking the StoredSafe REST-like API.

Because the wrapper is transparent, the only required return data
is the token on login requests and the HTTP status to indicate whether
the request was correct or not.
"""
MOCK_HOST = "test.storedsafe.com"
MOCK_URL = f"https://{MOCK_HOST}"
MOCK_URL_MTLS = f"https://{MOCK_HOST}:8443"
MOCK_TOKEN = "mock-token"
MOCK_APIKEY = "mock-api-key"
MOCK_USERNAME = "mock-username"
MOCK_PASSPHRASE = "mock-passphrase"
MOCK_OTP = "mock-otp"

MOCK_VAULT_ID = 42
MOCK_OBJECT_ID = 93
MOCK_TEMPLATE_ID = 10
MOCK_USER_ID = 18
MOCK_STATUS = 2
MOCK_NEEDLE = 'Waldo'
MOCK_SEARCH_STRING = 'Nilsson'

MOCK_PUBKEY = '/mock/certs/cert.pem'
MOCK_PRIVKEY = '/mock/certs/cert.key'

MOCK_PARAMS = {'first': 1, 'second': 'two', 'third': True}

MOCK_DEFAULT_OPTIONS = {'option_always': 1,
                        'option_override': 1,
                        'headers': {'x-http-default': 1}}
MOCK_OVERRIDE_OPTIONS = {'option_override': 2,
                         'headers': {'x-http-override': 2}}

MOCK_FILE_SIZE = 128
MOCK_FILE_EXTENSION = ".txt"
MOCK_FILE_CONTENT = b"A"*MOCK_FILE_SIZE
MOCK_FILE_PARAMS = { **MOCK_PARAMS, 'templateid': 3 }


class MockResponse:
    """
    Mimicks a minimal StoredSafe response as presented by the
    requests library.
    """

    def __init__(self, json_data, status_code, **kwargs):
        self.json_data = json_data
        self.status_code = status_code
        self.options = kwargs

    def json(self):
        """Returns the JSON data of the response."""
        return self.json_data


class MockError(MockResponse):
    """MockResponse prepared with error data"""

    def __init__(self, **kwargs):
        super().__init__({}, 403, **kwargs)

class MockNotFound(MockResponse):
    """MockResponse prepared with not found (404) data"""

    def __init__(self, **kwargs):
        super().__init__({}, 404, **kwargs)


class MockLoginSuccess(MockResponse):
    """MockResponse prepared with login data"""

    def __init__(self, **kwargs):
        super().__init__({'CALLINFO': {'token': MOCK_TOKEN}}, 200, **kwargs)


class MockSuccess(MockResponse):
    """MockResponse prepared with success data"""

    def __init__(self, **kwargs):
        super().__init__({}, 200, **kwargs)


def is_endpoint(endpoint, url, mtls=False):
    """Check if the intended enpoint matches the URL"""
    if mtls:
        return url == MOCK_URL_MTLS + endpoint
    return url == MOCK_URL + endpoint


def has_valid_token(**kwargs):
    """Check if a valid token exists in the header"""
    headers = kwargs.get('headers', {})
    return headers.get('X-Http-Token') == MOCK_TOKEN


def has_merged_options(options):
    headers = options.get('headers', {})
    return\
        options.get('option_always') == 1 and\
        options.get('option_override') == 2 and\
        headers.get('x-http-default') == 1 and\
        headers.get('x-http-override') == 2



"""
Test that /template endpoints use the correct format.
"""
import unittest
from unittest.mock import patch
from storedsafe import StoredSafe
# pylint: disable=unused-wildcard-import,wildcard-import
from mocks import *


class Template(unittest.TestCase):
    """Test /template endpoint"""

    def setUp(self):
        self.api = StoredSafe(
            host=MOCK_HOST, token=MOCK_TOKEN, **MOCK_DEFAULT_OPTIONS)

    @patch('requests.get', list_templates)
    def test_list_templates(self):
        """Successfully list templates"""
        res = self.api.list_templates(requests_options=MOCK_OVERRIDE_OPTIONS)
        self.assertEqual(res.status_code, 200)
        self.assertTrue(has_merged_options(res.options), True)

    @patch('requests.get', get_template)
    def test_get_template(self):
        """Successfully get template"""
        res = self.api.get_template(
            MOCK_TEMPLATE_ID, requests_options=MOCK_OVERRIDE_OPTIONS)
        self.assertEqual(res.status_code, 200)
        self.assertTrue(has_merged_options(res.options), True)



"""
/object endpoint mocks
"""
# pylint: disable=unused-wildcard-import,wildcard-import
from .mock_response import *


def list_users(*args, **kwargs):
    """Send mocked response listing all users."""
    if is_endpoint('/api/1.0/user', args[0]):
        if has_valid_token(**kwargs):
            return MockSuccess(**kwargs)
        return MockError(**kwargs)
    return MockNotFound(**kwargs)


def list_users_search_string(*args, **kwargs):
    """Send mocked response listing all users matching search string."""
    if is_endpoint(f'/api/1.0/user/{MOCK_SEARCH_STRING}', args[0]):
        if has_valid_token(**kwargs):
            return MockSuccess(**kwargs)
        return MockError(**kwargs)
    return MockNotFound(**kwargs)


def get_user(*args, **kwargs):
    """Send mocked response getting user."""
    if is_endpoint(f'/api/1.0/user/{MOCK_USER_ID}', args[0]):
        if has_valid_token(**kwargs):
            return MockSuccess(**kwargs)
        return MockError(**kwargs)
    return MockNotFound(**kwargs)


def create_user(*args, **kwargs):
    """Send mocked response creating user."""
    data = kwargs.get('json', {})
    if is_endpoint('/api/1.0/user', args[0]):
        if has_valid_token(**kwargs) and data == MOCK_PARAMS:
            return MockSuccess(**kwargs)
        return MockError(**kwargs)
    return MockNotFound(**kwargs)


def edit_user(*args, **kwargs):
    """Send mocked response editing user."""
    data = kwargs.get('json', {})
    if is_endpoint(f'/api/1.0/user/{MOCK_USER_ID}', args[0]):
        if has_valid_token(**kwargs) and data == MOCK_PARAMS:
            return MockSuccess(**kwargs)
        return MockError(**kwargs)
    return MockNotFound(**kwargs)


def delete_user(*args, **kwargs):
    """Send mocked response editing user."""
    if is_endpoint(f'/api/1.0/user/{MOCK_USER_ID}', args[0]):
        if has_valid_token(**kwargs):
            return MockSuccess(**kwargs)
        return MockError(**kwargs)
    return MockNotFound(**kwargs)
